//! Error handling stuffs
use std::{
    error,
    fmt,
};

pub trait Context
{
    fn line(&self) -> Option<u32>;
    fn column(&self) -> Option<u32>;
    fn file(&self) -> Option<&str>;
}

pub trait Reporter
{
    type Error: error::Error + ?Sized;

    #[inline] fn context(&self) -> Option<&dyn Context>
    {
	None
    }
    fn source(&self) -> &Self::Error;
    
    fn short_message(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result;
    #[inline] fn long_message(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	self.short_message(f)
    }
}

pub struct ShortMessage<R>(R)
where R: Reporter;

impl<R> fmt::Display for ShortMessage<R>
where R: Reporter
{
    #[inline] fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	self.0.short_message(f)
    }
}

pub struct LongMessage<R>(R)
where R: Reporter;

impl<R> fmt::Display for LongMessage<R>
where R: Reporter
{
    #[inline] fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	self.0.long_message(f)
    }
}

impl<E> Reporter for Box<E>
where E: error::Error + ?Sized
{
    type Error = E;

    fn source(&self) -> &Self::Error
    {
	&self
    }
    
    fn short_message(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	write!(f, "{}", &self)
    }
}

impl<'a, R> Reporter for &'a R
where R: Reporter
{
    type Error = <R as Reporter>::Error;
    
    #[inline] fn context(&self) -> Option<&dyn Context>
    {
	(*self).context()
    }
    fn source(&self) -> &Self::Error
    {
	(*self).source()
    }
    
    fn short_message(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	(*self).short_message(f)
    }
    #[inline] fn long_message(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	(*self).long_message(f)
    }
}

pub struct Wrap<E,M>(E, M);

impl<E,M> Reporter for Wrap<E,M>
where E: Reporter,
M: fmt::Display
{
    type Error = E::Error;

    #[inline] fn context(&self) -> Option<&dyn Context>
    {
	self.0.context()
    }
    fn source(&self) -> &Self::Error
    {
	self.0.source()
    }
    
    fn short_message(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	write!(f, "{}", self.1)
    }
    #[inline] fn long_message(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	write!(f, "{}", self.1)?;
	self.0.long_message(f) //ehh, also failed
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    fn test() -> Box<dyn error::Error>
    {
	Box::new(std::io::Error::new(std::io::ErrorKind::UnexpectedEof, "uhhhhh"))
    }
    #[test]
    fn atest()
    {
	let bx = test();
	assert_eq!(format!("{}", LongMessage(&bx)), format!("{}", bx));
    }
}
